# mypy: allow-untyped-defs
r"""Implementation for Stochastic Gradient Descent optimizer."""
from typing import List, Optional

import torch
from torch import Tensor
from corerec.torch_utils._foreach_utils import _get_fused_kernels_supported_devices
from corerec.cr_boosters.optimizer import (
    _default_to_fused_or_foreach,
    _differentiable_doc,
    _foreach_doc,
    _fused_doc,
    _maximize_doc,
    _use_grad_for_differentiable,
    DeviceDict,
    Optimizer,
)

__all__ = ["SGD", "sgd"]


class SGD(Optimizer):  # noqa: D101
    def __init__(
        self,
        params,
        lr: float = 1e-3,
        momentum: float = 0,
        dampening: float = 0,
        weight_decay: float = 0,
        nesterov=False,
        *,
        maximize: bool = False,
        foreach: Optional[bool] = None,
        differentiable: bool = False,
        fused: Optional[bool] = None,
    ):  # noqa: D107
        if lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if momentum < 0.0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        if weight_decay < 0.0:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        defaults = dict(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov,
            maximize=maximize,
            foreach=foreach,
            differentiable=differentiable,
            fused=fused,
        )
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super().__init__(params, defaults)

        if fused:
            self._step_supports_amp_scaling = True

            fused_supported_devices = _get_fused_kernels_supported_devices()
            if not all(
                p.device.type in fused_supported_devices and torch.is_floating_point(p)
                for pg in self.param_groups
                for p in pg["params"]
            ):
                raise RuntimeError(
                    "`fused=True` requires all the params to be floating point Tensors of "
                    f"supported devices: {fused_supported_devices}."
                )
            if differentiable:
                raise RuntimeError("`fused` does not support `differentiable`")
            if foreach:
                raise RuntimeError("`fused` and `foreach` cannot be `True` together.")

    def __setstate__(self, state):  # noqa: D105
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("nesterov", False)
            group.setdefault("maximize", False)
            group.setdefault("foreach", None)
            group.setdefault("differentiable", False)
            group.setdefault("fused", False)

    def _init_group(self, group, params, grads, momentum_buffer_list):
        has_sparse_grad = False

        for p in group["params"]:
            if p.grad is not None:
                params.append(p)
                grads.append(p.grad)
                if p.grad.is_sparse:
                    has_sparse_grad = True

                if group["momentum"] != 0:
                    state = self.state[p]
                    momentum_buffer_list.append(state.get("momentum_buffer"))

        return has_sparse_grad

    @_use_grad_for_differentiable
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (Callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params: List[Tensor] = []
            grads: List[Tensor] = []
            momentum_buffer_list: List[Optional[Tensor]] = []

            has_sparse_grad = self._init_group(
                group, params, grads, momentum_buffer_list
            )

            sgd(
                params,
                grads,
                momentum_buffer_list,
                weight_decay=group["weight_decay"],
                momentum=group["momentum"],
                lr=group["lr"],
                dampening=group["dampening"],
                nesterov=group["nesterov"],
                maximize=group["maximize"],
                has_sparse_grad=has_sparse_grad,
                foreach=group["foreach"],
                fused=group["fused"],
                grad_scale=getattr(self, "grad_scale", None),
                found_inf=getattr(self, "found_inf", None),
            )

            if group["momentum"] != 0:
                # update momentum_buffers in state
                for p, momentum_buffer in zip(params, momentum_buffer_list):
                    state = self.state[p]
                    state["momentum_buffer"] = momentum_buffer

        return loss


SGD.__doc__ = (
    r"""Implements stochastic gradient descent (optionally with momentum).

    .. math::
       \begin{aligned}
            &\rule{110mm}{0.4pt}                                                                 \\
            &\textbf{input}      : \gamma \text{ (lr)}, \: \theta_0 \text{ (params)}, \: f(\theta)
                \text{ (objective)}, \: \lambda \text{ (weight decay)},                          \\
            &\hspace{13mm} \:\mu \text{ (momentum)}, \:\tau \text{ (dampening)},
            \:\textit{ nesterov,}\:\textit{ maximize}                                     \\[-1.ex]
            &\rule{110mm}{0.4pt}                                                                 \\
            &\textbf{for} \: t=1 \: \textbf{to} \: \ldots \: \textbf{do}                         \\
            &\hspace{5mm}g_t           \leftarrow   \nabla_{\theta} f_t (\theta_{t-1})           \\
            &\hspace{5mm}\textbf{if} \: \lambda \neq 0                                           \\
            &\hspace{10mm} g_t \leftarrow g_t + \lambda  \theta_{t-1}                            \\
            &\hspace{5mm}\textbf{if} \: \mu \neq 0                                               \\
            &\hspace{10mm}\textbf{if} \: t > 1                                                   \\
            &\hspace{15mm} \textbf{b}_t \leftarrow \mu \textbf{b}_{t-1} + (1-\tau) g_t           \\
            &\hspace{10mm}\textbf{else}                                                          \\
            &\hspace{15mm} \textbf{b}_t \leftarrow g_t                                           \\
            &\hspace{10mm}\textbf{if} \: \textit{nesterov}                                       \\
            &\hspace{15mm} g_t \leftarrow g_{t} + \mu \textbf{b}_t                             \\
            &\hspace{10mm}\textbf{else}                                                   \\[-1.ex]
            &\hspace{15mm} g_t  \leftarrow  \textbf{b}_t                                         \\
            &\hspace{5mm}\textbf{if} \: \textit{maximize}                                          \\
            &\hspace{10mm}\theta_t \leftarrow \theta_{t-1} + \gamma g_t                   \\[-1.ex]
            &\hspace{5mm}\textbf{else}                                                    \\[-1.ex]
            &\hspace{10mm}\theta_t \leftarrow \theta_{t-1} - \gamma g_t                   \\[-1.ex]
            &\rule{110mm}{0.4pt}                                                          \\[-1.ex]
            &\bf{return} \:  \theta_t                                                     \\[-1.ex]
            &\rule{110mm}{0.4pt}                                                          \\[-1.ex]
       \end{aligned}

    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.
    """
    + rf"""
    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
        {_maximize_doc}
        {_foreach_doc}
        {_differentiable_doc}
        {_fused_doc}
    """
    + r"""

    Example:
        >>> # xdoctest: +SKIP
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et al. and implementations in some other frameworks.

        Considering the specific case of Momentum, the update can be written as

        .. math::
            \begin{aligned}
                v_{t+1} & = \mu * v_{t} + g_{t+1}, \\
                p_{t+1} & = p_{t} - \text{lr} * v_{t+1},
            \end{aligned}

        where :math:`p`, :math:`g`, :math:`v` and :math:`\mu` denote the
        parameters, gradient, velocity, and momentum respectively.

        This is in contrast to Sutskever et al. and
        other frameworks which employ an update of the form

        .. math::
            \begin{aligned}
                v_{t+1} & = \mu * v_{t} + \text{lr} * g_{t+1}, \\
                p_{t+1} & = p_{t} - v_{t+1}.
            \end{aligned}

        The Nesterov version is analogously modified.

        Moreover, the initial value of the momentum buffer is set to the
        gradient value at the first step. This is in contrast to some other
        frameworks that initialize it to all zeros.

    """
)


def sgd(
    params: List[Tensor],
    d_p_list: List[Tensor],
    momentum_buffer_list: List[Optional[Tensor]],
    # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
    # setting this as kwarg for now as functional API is compiled by torch/distributed/optim
    has_sparse_grad: bool = False,
    foreach: Optional[bool] = None,
    fused: Optional[bool] = None,
    grad_scale: Optional[Tensor] = None,
    found_inf: Optional[Tensor] = None,
    *,
    weight_decay: float,
    momentum: float,
    lr: float,
    dampening: float,
    nesterov: bool,
    maximize: bool,
):
    r"""Functional API that performs SGD algorithm computation.

    See :class:`~torch.optim.SGD` for details.
    """
    # Respect when the user inputs False/True for foreach or fused. We only want to change
    # the default when neither have been user-specified. Note that we default to foreach
    # and pass False to use_fused. This is not a mistake--we want to give the fused impl
    # bake-in time before making it the default, even if it is typically faster.
    if foreach is None and fused is None:
        # why must we be explicit about an if statement for torch.jit.is_scripting here?
        # because JIT can't handle Optionals nor fancy conditionals when scripting
        if not torch.jit.is_scripting():
            fused, foreach = _default_to_fused_or_foreach(
                params, differentiable=False, use_fused=False
            )
        else:
            foreach = False
            fused = False
    if foreach is None:
        foreach = False
    if fused is None:
        fused = False

    if foreach and torch.jit.is_scripting():
        raise RuntimeError("torch.jit.script not supported with foreach optimizers")
    if fused and torch.jit.is_scripting():
        raise RuntimeError("torch.jit.script not supported with fused optimizers")

    if foreach and not torch.jit.is_scripting():
        func = _multi_tensor_sgd
    elif fused and not torch.jit.is_scripting():
        func = _fused_sgd
    else:
        func = _single_tensor_sgd

    func(
        params,
        d_p_list,
        momentum_buffer_list,
        weight_decay=weight_decay,
        momentum=momentum,
        lr=lr,
        dampening=dampening,
        nesterov=nesterov,
        has_sparse_grad=has_sparse_grad,
        maximize=maximize,
        grad_scale=grad_scale,
        found_inf=found_inf,
    )


def _single_tensor_sgd(
    params: List[Tensor],
    grads: List[Tensor],
    momentum_buffer_list: List[Optional[Tensor]],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    weight_decay: float,
    momentum: float,
    lr: float,
    dampening: float,
    nesterov: bool,
    maximize: bool,
    has_sparse_grad: bool,
):
    assert grad_scale is None and found_inf is None

    for i, param in enumerate(params):
        grad = grads[i] if not maximize else -grads[i]

        if weight_decay != 0:
            grad = grad.add(param, alpha=weight_decay)

        if momentum != 0:
            buf = momentum_buffer_list[i]

            if buf is None:
                buf = torch.clone(grad).detach()
                momentum_buffer_list[i] = buf
            else:
                buf.mul_(momentum).add_(grad, alpha=1 - dampening)

            if nesterov:
                grad = grad.add(buf, alpha=momentum)
            else:
                grad = buf

        param.add_(grad, alpha=-lr)


def _multi_tensor_sgd(
    params: List[Tensor],
    grads: List[Tensor],
    momentum_buffer_list: List[Optional[Tensor]],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    weight_decay: float,
    momentum: float,
    lr: float,
    dampening: float,
    nesterov: bool,
    maximize: bool,
    has_sparse_grad: bool,
):
    assert grad_scale is None and found_inf is None

    if len(params) == 0:
        return

    grouped_tensors = Optimizer._group_tensors_by_device_and_dtype(
        [params, grads, momentum_buffer_list], with_indices=True  # type: ignore[list-item]
    )
    for (
        device_params,
        device_grads,
        device_momentum_buffer_list,
    ), indices in grouped_tensors.values():
        device_has_sparse_grad = has_sparse_grad and any(
            grad.is_sparse for grad in device_grads
        )

        if maximize:
            device_grads = torch._foreach_neg(device_grads)  # type: ignore[assignment]

        if weight_decay != 0:
            # Re-use the intermediate memory (device_grads) already allocated for maximize
            if maximize:
                torch._foreach_add_(device_grads, device_params, alpha=weight_decay)
            else:
                device_grads = torch._foreach_add(  # type: ignore[assignment]
                    device_grads, device_params, alpha=weight_decay
                )

        if momentum != 0:
            bufs = []

            all_states_with_momentum_buffer = True
            for i in range(len(device_momentum_buffer_list)):
                if device_momentum_buffer_list[i] is None:
                    all_states_with_momentum_buffer = False
                    break
                else:
                    bufs.append(device_momentum_buffer_list[i])

            if all_states_with_momentum_buffer:
                torch._foreach_mul_(bufs, momentum)
                torch._foreach_add_(bufs, device_grads, alpha=1 - dampening)
            else:
                bufs = []
                for i in range(len(device_momentum_buffer_list)):
                    if device_momentum_buffer_list[i] is None:
                        buf = device_momentum_buffer_list[i] = momentum_buffer_list[
                            indices[i]
                        ] = torch.clone(device_grads[i]).detach()
                    else:
                        buf = device_momentum_buffer_list[i]
                        buf.mul_(momentum).add_(device_grads[i], alpha=1 - dampening)

                    bufs.append(buf)

            if nesterov:
                torch._foreach_add_(device_grads, bufs, alpha=momentum)
            else:
                device_grads = bufs

        if not device_has_sparse_grad:
            # handle internal item() call if lr is a tensor
            if isinstance(lr, torch.Tensor) and torch._utils.is_compiling():
                grads_x_lr = torch._foreach_mul(device_grads, -lr)
                torch._foreach_add_(device_params, grads_x_lr)
            else:
                torch._foreach_add_(device_params, device_grads, alpha=-lr)
        else:
            # foreach APIs don't support sparse
            for i in range(len(device_params)):
                device_params[i].add_(device_grads[i], alpha=-lr)


def _fused_sgd(
    params: List[Tensor],
    grads: List[Tensor],
    momentum_buffer_list: List[Optional[Tensor]],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    weight_decay: float,
    momentum: float,
    lr: float,
    dampening: float,
    nesterov: bool,
    maximize: bool,
    has_sparse_grad: bool,
) -> None:
    if not params:
        return
    if has_sparse_grad:
        raise RuntimeError("`_fused_sgd` does not support sparse gradients")
    grad_scale_dict: DeviceDict = (
        {grad_scale.device: grad_scale} if grad_scale is not None else {}
    )
    found_inf_dict: DeviceDict = (
        {found_inf.device: found_inf} if found_inf is not None else {}
    )

    no_momentum_buffer = momentum == 0
    is_first_step = (
        all(t is None for t in momentum_buffer_list) and not no_momentum_buffer
    )
    if is_first_step:
        for i, g in enumerate(grads):
            momentum_buffer_list[i] = torch.empty_like(g)
    grouped_tensors = Optimizer._group_tensors_by_device_and_dtype(
        [params, grads, momentum_buffer_list], with_indices=False  # type: ignore[list-item]
    )
    for (device, _), (
        (device_params, device_grads, device_momentum_buffer_list),
        _,
    ) in grouped_tensors.items():
        device_grad_scale, device_found_inf = None, None
        if grad_scale is not None:
            device_grad_scale = grad_scale_dict.setdefault(
                device, grad_scale.to(device)
            )
        if found_inf_dict is not None and found_inf is not None:
            device_found_inf = found_inf_dict.setdefault(device, found_inf.to(device))
        torch._fused_sgd_(
            device_params,
            device_grads,
            [] if no_momentum_buffer else device_momentum_buffer_list,
            weight_decay=weight_decay,
            momentum=momentum,
            lr=lr,
            dampening=dampening,
            nesterov=nesterov,
            maximize=maximize,
            is_first_step=is_first_step,
            grad_scale=device_grad_scale,
            found_inf=device_found_inf,
        )



import numpy as np
from scipy.sparse import csr_matrix
from typing import List, Optional, Dict, Any, Tuple
from ..base_recommender import BaseRecommender

class BPRBase(BaseRecommender):
    """
    Bayesian Personalized Ranking (BPR) for implicit feedback datasets.
    
    Implementation based on the paper:
    "BPR: Bayesian Personalized Ranking from Implicit Feedback" by Rendle et al.
    
    BPR optimizes for correct ranking of items using a pairwise loss function.
    It's specifically designed for implicit feedback scenarios where only positive
    interactions are observed.
    
    Parameters:
    -----------
    factors : int
        Number of latent factors
    learning_rate : float
        Learning rate for SGD
    regularization : float
        Regularization strength
    iterations : int
        Number of training iterations
    batch_size : int
        Size of mini-batches
    seed : Optional[int]
        Random seed for reproducibility
    """
    def __init__(
        self,
        factors: int = 100,
        learning_rate: float = 0.01,
        regularization: float = 0.01,
        iterations: int = 100,
        batch_size: int = 1000,
        seed: Optional[int] = None
    ):
        self.factors = factors
        self.learning_rate = learning_rate
        self.regularization = regularization
        self.iterations = iterations
        self.batch_size = batch_size
        self.seed = seed
        
        self.user_map = {}
        self.item_map = {}
        self.reverse_user_map = {}
        self.reverse_item_map = {}
        self.user_factors = None
        self.item_factors = None
        self.item_bias = None
    
    def _create_mappings(self, user_ids: List[int], item_ids: List[int]) -> None:
        """Create mappings between original IDs and matrix indices"""
        self.user_map = {user_id: idx for idx, user_id in enumerate(user_ids)}
        self.item_map = {item_id: idx for idx, item_id in enumerate(item_ids)}
        self.reverse_user_map = {idx: user_id for user_id, idx in self.user_map.items()}
        self.reverse_item_map = {idx: item_id for item_id, idx in self.item_map.items()}
    
    def _init_factors(self, n_users: int, n_items: int) -> None:
        """Initialize model parameters"""
        if self.seed is not None:
            np.random.seed(self.seed)
            
        # Initialize factors with small random values
        self.user_factors = np.random.normal(0, 0.1, (n_users, self.factors))
        self.item_factors = np.random.normal(0, 0.1, (n_items, self.factors))
        self.item_bias = np.zeros(n_items)
    
    def _sample_triplet(self, user_item_matrix: csr_matrix, n_items: int) -> Tuple[int, int, int]:
        """Sample a (user, positive_item, negative_item) triplet for training"""
        # Sample a random user that has at least one interaction
        user_idx = np.random.randint(0, user_item_matrix.shape[0])
        while user_item_matrix[user_idx].nnz == 0:
            user_idx = np.random.randint(0, user_item_matrix.shape[0])
        
        # Sample a positive item for this user
        pos_items = user_item_matrix[user_idx].indices
        pos_idx = pos_items[np.random.randint(0, len(pos_items))]
        
        # Sample a negative item for this user
        neg_idx = np.random.randint(0, n_items)
        while neg_idx in pos_items:
            neg_idx = np.random.randint(0, n_items)
        
        return user_idx, pos_idx, neg_idx
    
    def _compute_loss(self, user_idx: int, pos_idx: int, neg_idx: int) -> float:
        """Compute BPR loss for a single triplet"""
        # Compute scores
        pos_score = np.dot(self.user_factors[user_idx], self.item_factors[pos_idx]) + self.item_bias[pos_idx]
        neg_score = np.dot(self.user_factors[user_idx], self.item_factors[neg_idx]) + self.item_bias[neg_idx]
        
        # Compute loss: -log(sigmoid(pos_score - neg_score))
        loss = -np.log(1.0 / (1.0 + np.exp(-(pos_score - neg_score))))
        
        # Add regularization
        loss += self.regularization * (
            np.sum(self.user_factors[user_idx]**2) +
            np.sum(self.item_factors[pos_idx]**2) +
            np.sum(self.item_factors[neg_idx]**2) +
            self.item_bias[pos_idx]**2 +
            self.item_bias[neg_idx]**2
        )
        
        return loss
    
    def _update_factors(self, user_idx: int, pos_idx: int, neg_idx: int) -> None:
        """Update model parameters using SGD for a single triplet"""
        # Compute scores
        pos_score = np.dot(self.user_factors[user_idx], self.item_factors[pos_idx]) + self.item_bias[pos_idx]
        neg_score = np.dot(self.user_factors[user_idx], self.item_factors[neg_idx]) + self.item_bias[neg_idx]
        
        # Compute sigmoid gradient
        sigmoid = 1.0 / (1.0 + np.exp(pos_score - neg_score))
        
        # Compute gradients
        grad_user = sigmoid * (self.item_factors[neg_idx] - self.item_factors[pos_idx]) + \
                    self.regularization * self.user_factors[user_idx]
                    
        grad_pos_item = sigmoid * (-self.user_factors[user_idx]) + \
                        self.regularization * self.item_factors[pos_idx]
                        
        grad_neg_item = sigmoid * self.user_factors[user_idx] + \
                        self.regularization * self.item_factors[neg_idx]
                        
        grad_pos_bias = sigmoid * (-1) + self.regularization * self.item_bias[pos_idx]
        grad_neg_bias = sigmoid * 1 + self.regularization * self.item_bias[neg_idx]
        
        # Update parameters
        self.user_factors[user_idx] -= self.learning_rate * grad_user
        self.item_factors[pos_idx] -= self.learning_rate * grad_pos_item
        self.item_factors[neg_idx] -= self.learning_rate * grad_neg_item
        self.item_bias[pos_idx] -= self.learning_rate * grad_pos_bias
        self.item_bias[neg_idx] -= self.learning_rate * grad_neg_bias
    
    def fit(self, interaction_matrix: csr_matrix, user_ids: List[int], item_ids: List[int]) -> None:
        """
        Train the BPR model on the given interaction data.
        
        Parameters:
        -----------
        interaction_matrix : csr_matrix
            User-item interaction matrix where non-zero entries indicate interactions
        user_ids : List[int]
            List of user IDs corresponding to rows in the interaction matrix
        item_ids : List[int]
            List of item IDs corresponding to columns in the interaction matrix
        """
        # Create mappings
        self._create_mappings(user_ids, item_ids)
        
        # Store interaction matrix for later use
        self.user_item_matrix = interaction_matrix
        
        # Get dimensions
        n_users, n_items = interaction_matrix.shape
        
        # Initialize model parameters
        self._init_factors(n_users, n_items)
        
        # Training loop
        for iteration in range(self.iterations):
            total_loss = 0.0
            
            # Mini-batch training
            for _ in range(self.batch_size):
                # Sample a triplet
                user_idx, pos_idx, neg_idx = self._sample_triplet(interaction_matrix, n_items)
                
                # Compute loss
                loss = self._compute_loss(user_idx, pos_idx, neg_idx)
                total_loss += loss
                
                # Update factors
                self._update_factors(user_idx, pos_idx, neg_idx)
            
            # Print progress
            avg_loss = total_loss / self.batch_size
            if (iteration + 1) % 10 == 0 or iteration == 0:
                print(f"Iteration {iteration+1}/{self.iterations}, Loss: {avg_loss:.4f}")
    
    def recommend(self, user_id: int, top_n: int = 10, exclude_seen: bool = True) -> List[int]:
        """
        Generate top-N recommendations for a specific user.
        
        Parameters:
        -----------
        user_id : int
            ID of the user to generate recommendations for
        top_n : int
            Number of recommendations to generate
        exclude_seen : bool
            Whether to exclude items the user has already interacted with
            
        Returns:
        --------
        List[int] : List of recommended item IDs
        """
        if self.user_factors is None or self.item_factors is None:
            raise ValueError("Model has not been trained. Call fit() first.")
        
        # Map user_id to internal index
        if user_id not in self.user_map:
            raise ValueError(f"User ID {user_id} not found in training data")
            
        user_idx = self.user_map[user_id]
        
        # Calculate scores for all items
        user_vector = self.user_factors[user_idx]
        scores = np.dot(self.item_factors, user_vector) + self.item_bias
        
        # If requested, exclude items the user has already interacted with
        if exclude_seen:
            seen_items = self.user_item_matrix[user_idx].indices
            scores[seen_items] = float('-inf')
        
        # Get top-n item indices
        top_item_indices = np.argsort(-scores)[:top_n]
        
        # Map indices back to original item IDs
        top_items = [self.reverse_item_map[idx] for idx in top_item_indices]
        
        return top_items
    
    def save_model(self, filepath: str) -> None:
        """Save the model to a file"""
        if self.user_factors is None or self.item_factors is None:
            raise ValueError("Model has not been trained. Call fit() first.")
            
        # Save model data
        model_data = {
            'user_factors': self.user_factors,
            'item_factors': self.item_factors,
            'item_bias': self.item_bias,
            'user_map': self.user_map,
            'item_map': self.item_map,
            'reverse_user_map': self.reverse_user_map,
            'reverse_item_map': self.reverse_item_map,
            'params': {
                'factors': self.factors,
                'learning_rate': self.learning_rate,
                'regularization': self.regularization,
                'iterations': self.iterations,
                'batch_size': self.batch_size,
                'seed': self.seed
            }
        }
        np.save(filepath, model_data, allow_pickle=True)
    
    @classmethod
    def load_model(cls, filepath: str) -> 'BPRBase':
        """Load a model from a file"""
        model_data = np.load(filepath, allow_pickle=True).item()
        
        # Create an instance with the saved parameters
        instance = cls(
            factors=model_data['params']['factors'],
            learning_rate=model_data['params']['learning_rate'],
            regularization=model_data['params']['regularization'],
            iterations=model_data['params']['iterations'],
            batch_size=model_data['params']['batch_size'],
            seed=model_data['params']['seed']
        )
        
        # Restore instance variables
        instance.user_factors = model_data['user_factors']
        instance.item_factors = model_data['item_factors']
        instance.item_bias = model_data['item_bias']
        instance.user_map = model_data['user_map']
        instance.item_map = model_data['item_map']
        instance.reverse_user_map = model_data['reverse_user_map']
        instance.reverse_item_map = model_data['reverse_item_map']
        
        return instance 


# dkn.py
import torch
import corerec.torch_nn as nn  # Use custom module instead of torch.nn
import torch.nn.functional as F

def get_divisible_num_heads(embed_dim, max_heads=8):
    """
    Returns the largest number of heads less than or equal to max_heads that divides embed_dim.

    Args:
        embed_dim (int): The embedding dimension.
        max_heads (int): The maximum number of heads to consider.

    Returns:
        int: A suitable number of heads.
    """
    for heads in range(max_heads, 0, -1):
        if embed_dim % heads == 0:
            return heads
    return 1  # Fallback to single head if no suitable number found

class DKN(nn.Module):
    def __init__(self, 
                 vocab_size, 
                 embedding_dim, 
                 entity_embedding_dim, 
                 knowledge_graph_size, 
                 text_kernel_sizes=[3, 4, 5], 
                 text_num_filters=100, 
                 dropout=0.5,
                 num_classes=1):
        """
        Initialize the DKN model.

        Args:
            vocab_size (int): Size of the vocabulary for text encoding.
            embedding_dim (int): Dimension of word embeddings.
            entity_embedding_dim (int): Dimension of entity embeddings from the knowledge graph.
            knowledge_graph_size (int): Number of entities in the knowledge graph.
            text_kernel_sizes (list): List of kernel sizes for text CNN.
            text_num_filters (int): Number of filters per kernel size for text CNN.
            dropout (float): Dropout rate.
            num_classes (int): Number of output classes.
        """
        super(DKN, self).__init__()
        
        # Text Embedding Layer
        self.text_embedding = nn.Embedding(num_embeddings=vocab_size, embedding_dim=embedding_dim, padding_idx=0)
        
        # Text CNN Encoder
        self.text_convs = nn.ModuleList([
            nn.Conv1d(in_channels=embedding_dim, 
                      out_channels=text_num_filters, 
                      kernel_size=k)
            for k in text_kernel_sizes
        ])
        
        # Knowledge Graph Embedding Layer
        self.entity_embedding = nn.Embedding(num_embeddings=knowledge_graph_size, embedding_dim=entity_embedding_dim, padding_idx=0)
        
        # Attention Mechanism
        embed_dim = text_num_filters * len(text_kernel_sizes)  # 100 * 3 = 300
        num_heads = get_divisible_num_heads(embed_dim, max_heads=8)
        print(f"Initializing MultiheadAttention with embed_dim={embed_dim} and num_heads={num_heads}")
        assert embed_dim % num_heads == 0, "embed_dim must be divisible by num_heads"
        
        self.attention = nn.MultiheadAttention(embed_dim=embed_dim, num_heads=num_heads, dropout=0.1)
        
        # Fusion Layer
        self.fusion = nn.Linear(embed_dim + entity_embedding_dim, 256)
        
        # Dropout Layer
        self.dropout = nn.Dropout(dropout)
        
        # Output Layer
        self.fc_out = nn.Linear(256, num_classes)
        
    def forward(self, text, entities):
        """
        Forward pass of the DKN model.

        Args:
            text (torch.Tensor): Input text tensor of shape (batch_size, seq_length).
            entities (torch.Tensor): Input entity tensor of shape (batch_size, num_entities).

        Returns:
            torch.Tensor: Output logits of shape (batch_size, num_classes).
        """
        # Text Embedding
        embedded_text = self.text_embedding(text)  # (batch_size, seq_length, embedding_dim)
        embedded_text = embedded_text.permute(0, 2, 1)  # (batch_size, embedding_dim, seq_length)
        
        # Text CNN Encoder
        text_conv_out = [F.relu(conv(embedded_text)) for conv in self.text_convs]  # List of (batch_size, num_filters, L_out)
        text_pooled = [F.max_pool1d(conv, conv.size(2)).squeeze(2) for conv in text_conv_out]  # List of (batch_size, num_filters)
        text_features = torch.cat(text_pooled, dim=1)  # (batch_size, embed_dim)
        
        # Knowledge Graph Embedding
        entity_embedded = self.entity_embedding(entities)  # (batch_size, num_entities, entity_embedding_dim)
        entity_features = torch.mean(entity_embedded, dim=1)  # (batch_size, entity_embedding_dim)
        
        # Attention Mechanism
        text_features = text_features.unsqueeze(1)  # (batch_size, 1, embed_dim)
        text_features, _ = self.attention(text_features, text_features, text_features)  # (batch_size, 1, embed_dim)
        text_features = text_features.squeeze(1)  # (batch_size, embed_dim)
        
        # Fusion Layer
        fused = torch.cat([text_features, entity_features], dim=1)  # (batch_size, embed_dim + entity_embedding_dim)
        fused = self.fusion(fused)  # (batch_size, 256)
        fused = F.relu(fused)
        fused = self.dropout(fused)
        
        # Output Layer
        out = self.fc_out(fused)  # (batch_size, num_classes)
        
        return out
    


from .ALS_base import ALSBase as MF_ALS
from .A2SVD_base import A2SVDBase as MF_A2VSD
from .als_recommender import ALSRecommender as MF_ALS
from .bayesian_matrix_factorization_base import BayesianMatrixFactorizationBase as MF_BAYESIAN_MATRIX_FACTORIZATION
from .contextual_matrix_factorization_base import ContextualMatrixFactorizationBase as MF_CONTEXTUAL_MATRIX_FACTORIZATION
from .deep_matrix_factorization_base import DeepMatrixFactorizationBase as MF_DEEP_MATRIX_FACTORIZATION
from .factorization_machine_base import FactorizationMachineBase as MF_FACTORIZATION_MACHINES
from .hybrid_matrix_factorization_base import HybridMatrixFactorizationBase as MF_HYBRID_MATRIX_FACTORIZATION
from .hierarchical_poisson_factorization_base import HierarchicalPoissonFactorizationBase as MF_HIERARCHICAL_POISSON_FACTORIZATION
from .hybrid_regularization_matrix_factorization_base import HybridRegularizationMatrixFactorizationBase as MF_HYBRID_REGULARIZATION

from .incremental_matrix_factorization_base import IncrementalMatrixFactorizationBase as MF_INCREMENTAL_MATRIX_FACTORIZATION
from .Implicit_feedback_mf_base import ImplicitFeedbackMFBase as MF_IMPLICIT_FEEDBACK
from .kernelized_matrix_factorization_base import KernelizedMatrixFactorizationBase as MF_KERNELIZED_MATRIX_FACTORIZATION
from .neural_matrix_factorization_base import NeuralMatrixFactorizationBase as MF_NEURAL_MATRIX_FACTORIZATION

from .nmf_base import NMFBase as MF_NMF
from .pmf_base import PMFBase as MF_PMF

from .sgd_matrix_factorization_base import SGDMatrixFactorizationBase as MF_SGD_MATRIX_FACTORIZATION
from .svdpp_base import SVDPPBase as MF_SVDPP
from .svd_base import SVDBase as MF_SVD
from .temporal_matrix_factorization_base import TemporalMatrixFactorizationBase as MF_TEMPORAL_MATRIX_FACTORIZATION
from .user_based_uf import UserBasedUF as MF_USER_BASED_UF
from .weighted_matrix_factorization_base import WeightedMatrixFactorizationBase as MF_WEIGHTED_MATRIX_FACTORIZATION

__all__ = [
    "MF_ALS",
    "MF_A2VSD",
    "MF_ALS",
    "MF_BAYESIAN_MATRIX_FACTORIZATION",
    "MF_CONTEXTUAL_MATRIX_FACTORIZATION",
]


from .sasrec_base import SASRecBase as NN_SASREC
from .rnn_sequential_recommendation_base import RNN_sequential_recommendation_base as NN_RNN_SEQUENTIAL_RECOMMENDATION
from .self_supervised_learning_cf_base import SelfSupervisedLearningCFBase as NN_SELF_SUPERVISED_LEARNING_CF
from .xdeepfm_base import XDeepFMBase as NN_XDEEPFM
from .ncf_base import NCF_base as NN_NCF
from .autoencoder_cf_base import AutoencoderCFBase as NN_AUTOENCODER_CF
from .hybrid_deep_learning_base import HybridDeepLearningBase as NN_HYBRID_DEEP_LEARNING
from .nextitnet_base import NextItNetBase as NN_NEXTITNET
from .ssept_base import SSEPTBase as NN_SSEPT
from .Bert4Rec_base import Bert4Rec_base as NN_BERT4REC
from .gnn_ufilter_base import GNN_ufilter_base as NN_GNN
from .gan_ufilter_base import GAN_ufilter_base as NN_GAN
from .transformer_ufilter_base import TransformerUFilterBase as NN_TRANSFORMER
from .variational_autoencoder_ufilter_base import VariationalAutoencoderUFilterBase as NN_VARIATIONAL_AUTOENCODER
from .slirec_base import SLiRecBase as NN_SLIREC
from .SSR_base import SSR_base as NN_SSR
from .deep_mf_base import DeepMF_base as NN_DEEP_MF
from .bivae_base import BVAE_base as NN_BVAE
from .neural_mf_base import NeuralMFBase as NN_NEURAL_MF
from .FM_base import FM_base as NN_FM
from .DeepFM_base import DeepFM_base as NN_DEEPFM
from .PNN_base import PNN_base as NN_PNN
from .AFM_base import AFM_base as NN_AFM
from .BST_base import BST_base as NN_BST
from .NFM_base import NFM_base as NN_NFM
from .ENSFM_base import ENSFM_base as NN_ENSFM
from .DeepRec_base import DeepRec_base as NN_DEEPREC
from .DeepFEFM_base import DeepFEFM_base as NN_DEEPFEFM
from .GNN_base import GNN_base as NN_GNN_BASE
from .PLE_base import PLE_base as NN_PLE
from .gru_ufilter_base import GRU_ufilter_base as NN_GRU
from .FFM_base import FFM_base as NN_FFM
from .FGCNN_base import FGCNN_base as NN_FGCNN
from .neural_ufilter_base import NeuralCollaborativeFilteringBase as NN_NEURAL_CF
from .RALM_base import RALM_base as NN_RALM
from .AutoFI_base import AutoFI_base as NN_AUTOFI
from .ESMM_base import ESMM_base as NN_ESMM
from .ESCMM_base import ESCMM_base as NN_ESCMM
from .DMR_base import DMR_base as NN_DMR
from .caser_base import Caser_base as NN_CASER
from .ListWise_base import ListWise_base as NN_LISTWISE
from .DIFM_base import DIFM_base as NN_DIFM
from .DeepCrossing_base import DeepCrossing_base as NN_DEEPCROSSING
from .DCN_base import DCN_base as NN_DCN
from .MMOE_base import MMOE_base as NN_MMOE
from .Fibinet_base import Fibinet_base as NN_FIBINET
from .DLRM_base import DLRM_base as NN_DLRM
from .AutoInt_base import AutoInt_base as NN_AUTOINTER
from .MAML_base import MAML_base as NN_MAML
from .GateNet_base import GateNet_base as NN_GATENET
from .DIEN_base import DIEN_base as NN_DIEN
from .FLEN_base import FLEN_base as NN_FLEN
from .TiSAS_base import TiSASBase as NN_TISAS

__all__ = [
    "NN_SASREC",
    "NN_RNN_SEQUENTIAL_RECOMMENDATION",
    "NN_SELF_SUPERVISED_LEARNING_CF",
    "NN_XDEEPFM",
    "NN_NCF",
    "NN_AUTOENCODER_CF",
    "NN_HYBRID_DEEP_LEARNING",
    "NN_NEXTITNET",
    "NN_SSEPT",
    "NN_BERT4REC",
    "NN_GNN",
    "NN_GAN",
    "NN_TRANSFORMER",
    "NN_VARIATIONAL_AUTOENCODER",
    "NN_SLIREC",
    "NN_SSR",
    "NN_DEEP_MF",
    "NN_BVAE",
    "NN_NEURAL_MF",
    "NN_FM",
    "NN_DEEPFM",
    "NN_PNN",
    "NN_AFM",
    "NN_BST",
    "NN_NFM",
    "NN_ENSFM",
    "NN_DEEPREC",
    "NN_DEEPFEFM",
    "NN_GNN_BASE",
    "NN_PLE",
    "NN_GRU",
    "NN_FFM",
    "NN_FGCNN",
    "NN_NEURAL_CF",
    "NN_RALM",
    "NN_AUTOFI",
    "NN_ESMM",
    "NN_ESCMM",
    "NN_DMR",
    "NN_CASER",
    "NN_LISTWISE",
    "NN_DIFM",
    "NN_DEEPCROSSING",
    "NN_DCN",
    "NN_MMOE",
    "NN_FIBINET",
    "NN_DLRM",
    "NN_AUTOINTER",
    "NN_MAML",
    "NN_GATENET",
    "NN_DIEN",
    "NN_FLEN",
    "NN_DIN",
    "NN_TISAS",
]

# init is in maintenance mode
            


# fairness_aware implementation
import pandas as pd
from typing import Dict, List

class FAIRNESS_AWARE:
    def __init__(self):
        """
        Initialize the fairness-aware module.

        Attributes:
            fairness_metrics (dict): A dictionary to store calculated fairness metrics, 
            such as the distribution of recommendations across different user demographics.
        """
        self.fairness_metrics = {}

    def evaluate_fairness(self, recommendations: Dict[int, List[int]], user_attributes: pd.DataFrame) -> Dict[str, float]:
        """
        Evaluate the fairness of the recommendations across different user groups.

        Parameters:
            recommendations (dict): A dictionary mapping user IDs to lists of recommended item IDs.
            user_attributes (pd.DataFrame): A DataFrame containing user demographic information, 
            such as age, gender, and other relevant attributes.

        Returns:
            dict: A dictionary of fairness metrics, providing insights into how recommendations 
            are distributed across different user groups. For example, it may include the 
            distribution of recommendations by gender or age group.
        """
        # Example: Calculate the distribution of recommendations across gender
        gender_distribution = user_attributes['gender'].value_counts(normalize=True).to_dict()
        self.fairness_metrics['gender_distribution'] = gender_distribution
        return self.fairness_metrics

    def ensure_fairness(self, recommendations: Dict[int, List[int]], user_attributes: pd.DataFrame) -> Dict[int, List[int]]:
        """
        Adjust recommendations to ensure fairness across user groups.

        Parameters:
            recommendations (dict): A dictionary mapping user IDs to lists of recommended item IDs.
            user_attributes (pd.DataFrame): A DataFrame containing user demographic information, 
            such as age, gender, and other relevant attributes.

        Returns:
            dict: Adjusted recommendations ensuring fairness, potentially modifying the original 
            recommendations to achieve a more balanced distribution across user groups.
        """
        # Placeholder: Implement logic to adjust recommendations for fairness
        return recommendations



import numpy as np
from scipy.sparse import csr_matrix
from typing import List, Optional, Dict, Any, Tuple
from ..base_recommender import BaseRecommender

class FactorizationMachineBase(BaseRecommender):
    """
    Factorization Machine (FM) for collaborative filtering.
    
    Based on the paper:
    "Factorization Machines" by Steffen Rendle.
    
    This implementation focuses on the 2-way FM model for collaborative filtering,
    which captures pairwise interactions between features.
    
    Parameters:
    -----------
    factors : int
        Number of latent factors
    learning_rate : float
        Learning rate for SGD
    regularization : float
        Regularization strength
    iterations : int
        Number of training iterations
    batch_size : int
        Size of mini-batches
    seed : Optional[int]
        Random seed for reproducibility
    """
    def __init__(
        self,
        factors: int = 10,
        learning_rate: float = 0.01,
        regularization: float = 0.01,
        iterations: int = 100,
        batch_size: int = 1000,
        seed: Optional[int] = None
    ):
        self.factors = factors
        self.learning_rate = learning_rate
        self.regularization = regularization
        self.iterations = iterations
        self.batch_size = batch_size
        self.seed = seed
        
        self.user_map = {}
        self.item_map = {}
        self.reverse_user_map = {}
        self.reverse_item_map = {}
        self.w0 = None  # Global bias
        self.w = None   # Feature weights
        self.v = None   # Feature factor matrix
    
    def _create_mappings(self, user_ids: List[int], item_ids: List[int]) -> None:
        """Create mappings between original IDs and matrix indices"""
        self.user_map = {user_id: idx for idx, user_id in enumerate(user_ids)}
        self.item_map = {item_id: idx for idx, item_id in enumerate(item_ids)}
        self.reverse_user_map = {idx: user_id for user_id, idx in self.user_map.items()}
        self.reverse_item_map = {idx: item_id for item_id, idx in self.item_map.items()}
    
    def _init_params(self, n_users: int, n_items: int) -> None:
        """Initialize model parameters"""
        if self.seed is not None:
            np.random.seed(self.seed)
        
        # Number of features: user_id + item_id
        n_features = n_users + n_items
        
        # Initialize parameters
        self.w0 = 0.0  # Global bias
        self.w = np.zeros(n_features)  # Linear terms
        self.v = np.random.normal(0, 0.1, (n_features, self.factors))  # Interaction factors
    
    def _create_feature_vector(self, user_idx: int, item_idx: int, n_users: int) -> np.ndarray:
        """Create one-hot encoded feature vector for user-item pair"""
        # Feature vector: [user_1, user_2, ..., item_1, item_2, ...]
        x = np.zeros(n_users + self.user_item_matrix.shape[1])
        x[user_idx] = 1.0  # One-hot encode user
        x[n_users + item_idx] = 1.0  # One-hot encode item
        return x
    
    def _predict(self, x: np.ndarray) -> float:
        """Make prediction using FM model for a feature vector"""
        # First order term: w0 + sum(w_i * x_i)
        pred = self.w0 + np.dot(self.w, x)
        
        # Second order term: sum_f( (sum_i(v_i,f * x_i))^2 - sum_i(v_i,f^2 * x_i^2) ) / 2
        sum_square = np.zeros(self.factors)
        square_sum = np.zeros(self.factors)
        
        # Calculate interaction terms efficiently
        for i in range(len(x)):
            if x[i] != 0:
                sum_square += self.v[i] * x[i]
                square_sum += (self.v[i] * x[i]) ** 2
        
        # Add interaction term to prediction
        pred += 0.5 * np.sum(sum_square ** 2 - square_sum)
        
        return pred
    
    def _sgd_update(self, x: np.ndarray, y: float, pred: float) -> None:
        """Update parameters using SGD for a single sample"""
        # Calculate error
        error = pred - y
        
        # Update global bias
        self.w0 -= self.learning_rate * (error + self.regularization * self.w0)
        
        # Update feature weights and factors
        for i in range(len(x)):
            if x[i] == 0:
                continue
                
            # Update linear term
            grad_w = error * x[i] + self.regularization * self.w[i]
            self.w[i] -= self.learning_rate * grad_w
            
            # Calculate sum term for interaction factors
            sum_term = np.zeros(self.factors)
            for j in range(len(x)):
                if x[j] != 0 and j != i:
                    sum_term += self.v[j] * x[j]
            
            # Update interaction factors
            grad_v = error * x[i] * sum_term + self.regularization * self.v[i]
            self.v[i] -= self.learning_rate * grad_v
    
    def fit(self, interaction_matrix: csr_matrix, user_ids: List[int], item_ids: List[int]) -> None:
        """
        Train the FM model on the given interaction data.
        
        Parameters:
        -----------
        interaction_matrix : csr_matrix
            User-item interaction matrix where non-zero entries indicate interactions
        user_ids : List[int]
            List of user IDs corresponding to rows in the interaction matrix
        item_ids : List[int]
            List of item IDs corresponding to columns in the interaction matrix
        """
        # Create mappings
        self._create_mappings(user_ids, item_ids)
        
        # Store interaction matrix for later use
        self.user_item_matrix = interaction_matrix
        
        # Get dimensions
        n_users, n_items = interaction_matrix.shape
        
        # Initialize model parameters
        self._init_params(n_users, n_items)
        
        # Create training data
        user_indices, item_indices, ratings = [], [], []
        for user_idx in range(n_users):
            for item_idx in interaction_matrix[user_idx].indices:
                user_indices.append(user_idx)
                item_indices.append(item_idx)
                ratings.append(1.0)  # Assuming binary interactions
        
        # Training loop
        n_samples = len(user_indices)
        for iteration in range(self.iterations):
            # Shuffle training data
            indices = np.random.permutation(n_samples)
            user_indices_shuffled = [user_indices[i] for i in indices]
            item_indices_shuffled = [item_indices[i] for i in indices]
            ratings_shuffled = [ratings[i] for i in indices]
            
            total_loss = 0.0
            
            # Mini-batch training
            for batch_start in range(0, n_samples, self.batch_size):
                batch_end = min(batch_start + self.batch_size, n_samples)
                
                batch_loss = 0.0
                
                # Process each sample in the batch
                for i in range(batch_start, batch_end):
                    user_idx = user_indices_shuffled[i]
                    item_idx = item_indices_shuffled[i]
                    rating = ratings_shuffled[i]
                    
                    # Create feature vector
                    x = self._create_feature_vector(user_idx, item_idx, n_users)
                    
                    # Make prediction
                    pred = self._predict(x)
                    
                    # Calculate loss
                    error = pred - rating
                    batch_loss += error ** 2
                    
                    # Update parameters
                    self._sgd_update(x, rating, pred)
                
                total_loss += batch_loss
            
            # Print progress
            avg_loss = total_loss / n_samples
            if (iteration + 1) % 10 == 0 or iteration == 0:
                print(f"Iteration {iteration+1}/{self.iterations}, Loss: {avg_loss:.4f}")
    
    def recommend(self, user_id: int, top_n: int = 10, exclude_seen: bool = True) -> List[int]:
        """
        Generate top-N recommendations for a specific user.
        
        Parameters:
        -----------
        user_id : int
            ID of the user to generate recommendations for
        top_n : int
            Number of recommendations to generate
        exclude_seen : bool
            Whether to exclude items the user has already interacted with
            
        Returns:
        --------
        List[int] : List of recommended item IDs
        """
        if self.w is None or self.v is None:
            raise ValueError("Model has not been trained. Call fit() first.")
        
        # Map user_id to internal index
        if user_id not in self.user_map:
            raise ValueError(f"User ID {user_id} not found in training data")
            
        user_idx = self.user_map[user_id]
        n_users = len(self.user_map)
        n_items = len(self.item_map)
        
        # Calculate scores for all items
        scores = np.zeros(n_items)
        for item_idx in range(n_items):
            # Create feature vector
            x = self._create_feature_vector(user_idx, item_idx, n_users)
            
            # Make prediction
            scores[item_idx] = self._predict(x)
        
        # If requested, exclude items the user has already interacted with
        if exclude_seen:
            seen_items = self.user_item_matrix[user_idx].indices
            scores[seen_items] = float('-inf')
        
        # Get top-n item indices
        top_item_indices = np.argsort(-scores)[:top_n]
        
        # Map indices back to original item IDs
        top_items = [self.reverse_item_map[idx] for idx in top_item_indices]
        
        return top_items
    
    def save_model(self, filepath: str) -> None:
        """Save the model to a file"""
        if self.w is None or self.v is None:
            raise ValueError("Model has not been trained. Call fit() first.")
            
        # Save model data
        model_data = {
            'w0': self.w0,
            'w': self.w,
            'v': self.v,
            'user_map': self.user_map,
            'item_map': self.item_map,
            'reverse_user_map': self.reverse_user_map,
            'reverse_item_map': self.reverse_item_map,
            'params': {
                'factors': self.factors,
                'learning_rate': self.learning_rate,
                'regularization': self.regularization,
                'iterations': self.iterations,
                'batch_size': self.batch_size,
                'seed': self.seed
            }
        }
        np.save(filepath, model_data, allow_pickle=True)
    
    @classmethod
    def load_model(cls, filepath: str) -> 'FactorizationMachineBase':
        """Load a model from a file"""
        model_data = np.load(filepath, allow_pickle=True).item()
        
        # Create an instance with the saved parameters
        instance = cls(
            factors=model_data['params']['factors'],
            learning_rate=model_data['params']['learning_rate'],
            regularization=model_data['params']['regularization'],
            iterations=model_data['params']['iterations'],
            batch_size=model_data['params']['batch_size'],
            seed=model_data['params']['seed']
        )
        
        # Restore instance variables
        instance.w0 = model_data['w0']
        instance.w = model_data['w']
        instance.v = model_data['v']
        instance.user_map = model_data['user_map']
        instance.item_map = model_data['item_map']
        instance.reverse_user_map = model_data['reverse_user_map']
        instance.reverse_item_map = model_data['reverse_item_map']
        
        return instance 


import math
from collections import defaultdict

class TFIDFExtractor:
    def __init__(self, text_column, max_features=1000):
        self.text_column = text_column
        self.max_features = max_features
        self.vocabulary = {}
        self.idf = {}

    def extract_features(self, data):
        # Calculate Term Frequency (TF)
        tf = []
        document_count = len(data)
        term_document_counts = defaultdict(int)

        for text in data[self.text_column]:
            terms = text.split('|')
            term_freq = defaultdict(int)
            unique_terms = set()
            for term in terms:
                term_freq[term] += 1
                unique_terms.add(term)
            tf.append(term_freq)
            for term in unique_terms:
                term_document_counts[term] += 1

        # Calculate Inverse Document Frequency (IDF)
        for term, count in term_document_counts.items():
            self.idf[term] = math.log((1 + document_count) / (1 + count)) + 1

        # Build Vocabulary
        all_terms = sorted(self.idf.keys(), key=lambda x: self.idf[x], reverse=True)
        self.vocabulary = {term: idx for idx, term in enumerate(all_terms[:self.max_features])}

        # Calculate TF-IDF
        tfidf_matrix = []
        for term_freq in tf:
            vector = [0.0] * len(self.vocabulary)
            for term, freq in term_freq.items():
                if term in self.vocabulary:
                    idx = self.vocabulary[term]
                    vector[idx] = freq * self.idf[term]
            tfidf_matrix.append(vector)

        return tfidf_matrix


"""
Ensemble Methods for Hybrid Recommendation Systems

This module implements ensemble methods that combine multiple recommendation models
to improve overall system performance. Ensemble methods leverage the diversity of
different models to enhance prediction accuracy and robustness.

Key Features:
- Supports various ensemble strategies, including bagging, boosting, and stacking.
- Combines outputs from multiple models to generate final recommendations.
- Provides flexibility in model selection and ensemble configuration.

Classes:
- ENSEMBLE_METHODS: Main class implementing ensemble techniques for hybrid
  recommendation systems.

Usage:
Create an instance of the ENSEMBLE_METHODS class to configure and apply ensemble
techniques to your recommendation models. Use the provided methods to train and
generate ensemble-based recommendations.

Example:
    ensemble_model = ENSEMBLE_METHODS()
    ensemble_model.train(models_list, user_item_matrix)
    final_recommendations = ensemble_model.recommend(user_id, top_n=10)
"""
pass
